#!/usr/local/bin/perl

use Config;
use File::Basename qw(&basename &dirname);

# List explicitly here the variables you want Configure to
# generate.  Metaconfig only looks for shell variables, so you
# have to mention them as if they were shell variables, not
# %Config entries.  Thus you write
#  $startperl
# to ensure Configure will look for $Config{startperl}.

# This forces PL files to create target in same directory as PL file.
# This is so that make depend always knows where to find PL derivatives.
chdir(dirname($0));
($file = basename($0)) =~ s/\.PL$//;
$file =~ s/\.pl$//
	if ($^O eq 'VMS' or $^O eq 'os2');  # "case-forgiving"

open OUT,">$file" or die "Can't create $file: $!";

print "Extracting $file (with variable substitutions)\n";

# In this section, perl variables will be expanded during extraction.
# You can use $Config{...} to use Configure variables.

print OUT <<"!GROK!THIS!";
$Config{'startperl'} 
!GROK!THIS!
print OUT <<'!NO!SUBS!';
my $RCS_Id = '$Id: despam.pl,v 2.8 1997-05-18 20:29:47+02 jv Exp $ ';

# Author          : Johan Vromans / Felix Ingrand
# Created On      : Tue Sep 15 15:59:04 1992
# Last Modified By: Johan Vromans
# Last Modified On: Sun May 18 20:28:41 1997
# Update Count    : 276
# Status          : Unknown, Use with caution!

################ Documentation ################

=head1 NAME

despam - filter email messages

=head1 SYNOPSIS

despam [options] -verify addresses...

despam [options] < email-message

B<IMPORTANT:> The despam program needs manual configuration before it
can be used!

=head1 OPTIONS

=over 8

=item B<-addresses> I<file>

Designates a file that contains a list of known email addresses, one
per line. Email messages from an address that is in this file will be
unconditionally accepted.

=item B<-data> I<file>

Designates a file that contains information about email addresses that
must be allowed or denied. See below.

=item B<-ignore_sender> I<sender>

Ignore this sender in a Sender: header. Necessary if you use a
remailer that always inserts a Sender: header with your own email
address. 

=item B<-domain> I<domain>

Accept email that has a References: or In-Reply-To: header that comes
from this domain.

=item B<-from> I<email address>

Insert a From: line in the reject mail with this contents. Useful if
you are using someone else's account to process your mail.

=item B<-tag> I<tag>

Add a unique tag text to the F<X-Despammed-By header>. This allows
bounced messages to be filtered using B<procmail>.

=item B<-key> I<key>

B<despam> allows a message to pass if it is re-submitted with a
special keyword (or F<X-Despam-Key> header. It checks if the message
was re-submitted within two weeks. I<key>, which must be a decimal
number, is used as a starting point for the calculations.

=item B<-check> I<addr>

Instead of sending the reject mail to the originator, sent it to the
given address. Useful to test your specific configuration before
putting it in production.

=item B<-nocache>

Do not cache the contents of the I<data> file. Useful for debugging.

=item B<-verify>

Do not process an email message, but verify addresses that are
specified on the command line. Use B<-verbose> also.

=item B<-help>

Print a brief help message and exits.

=item B<-man>

Prints the manual page and exits.

=item B<-ident>

Prints program identification.

=item B<-verbose>

More verbose information.

=back

=head1 DESCRIPTION

This program analyzes email messages, and filters them based on a
number of criteria. If a message does not match the criteria, the
message is considered a SPAM and send back to the sender(s).

Criteria are:

=over 4

=item *

The message contains a special X-Despam-Key: header with a correct
value (see below).

=item *

The message contains a special despamming key in the Subject:
header with a correct value (see below).

=item *

The Subject: header contains one or more buzzwords [Optional].

=item *

Of all the addresses from the From: Resent-From: Sender: and
Reply-To: headers there is at least one "known" address, and no
"denied" addresses.

=back

Addresses are considered "known" if

=over 4

=item *

the address is present in the email addresses file [Optional], or

=item *

the address is described in the I<allow> section of the I<data>
file [Optional].

=back

Addresses are considered "suspicious" or "denied" if

=over 4

=item *

the address is described in the corresponding section of the
I<data> file [Optional].

If the address is considered "denied" it is thrown away without
notice. 

All input files are optional. However, without data the program is
rather useless. 

The list of email addresses can be built by hand, but can be generated
from a BBDB file. A emacs-lisp program that comes with the despam kit
can be used to facilitate this.

=back

=head1 RETURN VALUES

=over 4

=item 0

Message passed.

=item 1

Commans syntax error.

=item 3

Message rejected, a notice was sent.

=item 4

Message rejected, no notice sent.

=back

=head1 Syntax of the data file

Empty lines and lines that start with "#" are ignored.

A section consists of a tag word on the beginning of the line,
followed by zero or more lines containing data. These lines must begin
with whitespace. Tag words may occur multiply, the data is
concatenated.

The data is taken literally and transformed into a Perl regular
expression pattern. Shell-type '*' wildcards can be used to denote
optional characters. The special sequence @@ can be used to denote
@[optional words.]. In general, this is used to match subdomains,
e.g. @@foo.com  will match @foo.com, but also @bar.foo.com, etc.
If an entry starts with a slash, it is assumed to be a perl
regular expression pattern and copied without modification.

=head2 buzz

This section contains buzzwords. If such a word (bounded by non-word
constituents) is found in the subject of a message, it is accepted.

Example:

    buzz
	re:
	perl

=head2 allow

Allowed addresses. An address that matches one of these is marked Ok.
If none of the other addresses are denied, the message is accepted.

Example:

    allow
	*@@perl.com
	*@franz.ww.tu-berlin.de

=head2 suspicious

Suspicious addresses. If any address matches any one of these, the
message is rejected and a notification is sent to the originator.

For example:

    suspicious
	*@@savetrees.com
	*@@microsoft.com

=head2 deny

If any address matches any one of these, the message is rejected; no
notification is sent.

=head1 Example

To use B<despam> with B<procmail>, install a script
F<$HOME/bin/despam> with contents:

    #! /bin/sh
    cd $HOME/lib/despam
    exec perl despam.pl --data=despam.dat --addresses=$HOME/.bbdb.email

In your F<.procmailrc>, use:

    # Despamming
    :0Wci: despam.lock
    |$HOME/bin/despam

    :0e:
    $HOME/Mail/Incoming/spam.spool

This causes every email message to be passed through the despamming
filter. If the message is rejected, it is stored in the F<spam.spool>
mailbox for later inspection.

To get rid of bounces (spam mails usually have incorrect addresses),
you can weed out mails that have the F<X-Despammed-By> header
somewhere in the body:

    :0B:
    * ^X-Despammed-By:
    $HOME/Mail/Incoming/spam.spool

=head1 AUTHORS

Basic idea from Johan Vromans E<lt>jvromans@squirrel.nl>. Lots of
improvements by Felix Ingrand E<lt>felix@laas.fr>. Then hacked again
by Johan Vromans.

Felix maintains a Web page about despam:
F<http://www.laas.fr/~felix/despam.html>. 

The authors do not accept any responsibility for consequences of using
this program.

=cut

################ Common stuff ################

use strict;
my $my_package = 'Sciurix';
my ($my_name, $my_version) = $RCS_Id =~ /: (.+).pl,v ([\d.]+)/;
$my_version .= '*' if length('$Locker:  $ ') > 12;

################ Program parameters ################

use Getopt::Long 2.00;
my $verbose = 0;

# Default list of email addresses
my $addresses = '';

# Default data file.
my $datafile = '';

# Relinquish rules for email from this domain (optional).
my $domain = '';

# Despamming tag.
my $despam_tag = '';

# Cache data file(s).
my $cache = 1;

# Verify only?
my $verify = 0;

# Use this a From: address for the reject mail.
my $from = '';

# Send reject to this address.
my $check = '';

# Ignore this address in Sender: headers.
my $i_sender = '';

my ($debug, $trace, $test) = (0, 0, 0);
options ();
$domain = quotemeta ($domain) if $domain ne '';

################ Presets ################

use Mail::Address 1.09;
use Mail::Header;

my $offset = 2378957;		#### CHANGE THIS ####
my $grace = 14 * 24 * 3600;	# two weeks

my $TMPDIR = $ENV{'TMPDIR'} || '/usr/tmp';

################ The Process ################

my $line;			# input line being scanned
my $src;			# original input line
# NOTE: These have to be local since they're require-d.
use vars qw ($re_deny $re_suspicious $re_allow $re_buzz);
local $re_deny = '';		# regexp with denied addresses
local $re_suspicious = '';	# regexp with suspicious addresses
local $re_allow = '';		# regexp with allowed addresses
local $re_buzz = '';		# regexp with buzz words
my $addrsfile;			# addrsses file, if opened

my @reason = ("Known", "Allowed", "Unknown", "Suspicious", "Denied");

my $ret;

if ( $verify ) {
    foreach $a ( @ARGV ) {
	$ret = vrfy_addr ($a);
	if ( $verbose ) {
	    print ($a, ": ", $reason[$ret], ".\n");
	}
    }
    exit $ret;
}

$ret = despam ();

exit $ret;

################ Subroutines ################

sub grep_addrsfile {
    my ($addr) = (lc(shift));

    unless ( defined $addrsfile ) {
	use Symbol;
	print STDERR ("=> opening $addresses\n") if $debug;
	$addrsfile = gensym;
	open ($addrsfile, $addresses) or die ("$addresses: $!\n");
    }
    else {
	seek ($addrsfile, 0, 0);
    }

    while ( <$addrsfile> ) {
	chop;
	return 1 if $addr eq lc($_);
    }
    return 0;
};

sub vrfy_addr {
    my ($addr) = @_;
    my $ret;
    print STDERR ("=> vrfy: $addr\n") if $debug;

    # Check for a known address in the address list.
    if ( $addresses ne '') {
	$ret = 0 if grep_addrsfile ($addr);
    }

    # Check for denied/suspicious/allowed addresses.
    if ( !defined $ret and $datafile ne "" ) {
	parse_data (0) if $re_buzz eq '';
	$ret = 4 if                   $addr =~ /^$re_deny$/i;
	$ret = 3 if !defined $ret and $addr =~ /^$re_suspicious$/i;
	$ret = 1 if !defined $ret and $addr =~ /^$re_allow$/i;
    }
    # Unknown.
    $ret = 2 unless defined $ret;

    print STDERR ("=> verified: $ret [$reason[$ret]]\n") if $debug;
    $ret;
}

# Despamming.
# Returns:
#	0: Passed
#	3: Rejected with notice
#	4: Rejected without notice
sub despam {
    
    my $orig;			# original headers

    # Read headers in paragraph mode.
    { local ($/) = ""; $orig = <>; }

    # Strip leading "From " line, if present.
    $orig =~ s/^From .*\n//;

    # Normalize headers. Join continuation lines.
    my $headers;
    ($headers = $orig) =~ s/[ \t]*\n[ \t]+/ /g;

    # Split into lines.
    my @headers = split (/\n/, $headers);

    # Parse.
    my $head = new Mail::Header(\@headers);

    # Check for de-despamming key.
    if ( $head->count("X-Despam-Key") ) {
	my $subj = $head->get("X-Despam-Key");
	if ( $subj =~ /Z([0-9A-F]{8})V/ ) {
	    my $try = hex($1) + $offset;
	    if ( time - $try > 0 and time - $try <= $grace) {
		notice ("Passed: Valid X-Despam-Key header");
		return 0;
	    }
      }
    }

    if ( $domain ne '' ) {
	# Check for In-Reply-To matching my domain.
	if ( $head->count("In-Reply-To") ) {
	    my $ref = $head->get("In-Reply-To");
	    if ( $ref =~ /\b$domain\b/i ) {
		notice ("Passed: In-Reply-To header matches my domain");
		return 0;
	    }
	}

	# Check for references to a mail from my domain.
	if ( $head->count("References") ) {
	    my $ref = $head->get("References");
	    if ( $ref =~ /\b$domain\b/i ) {
		notice ("Passed: References header matches my domain");
		return 0;
	    }
	}
    }

    # Check for despamming key or buzzwords in subject.
    if ( $head->count("Subject") ) {
	my $subj = $head->get("Subject");
	if ( $subj =~ /Z([0-9A-F]{8})V/ ) {
	    my $try = hex($1) + $offset;
	    if ( time - $try > 0 and time - $try <= $grace) {
		notice ("Passed: Valid de-despamming key in subject");
		return 0;
	    }
	}
	if ( $datafile ne '' and $re_buzz eq '' ) {
	    parse_data (0);
	}
	if ( $re_buzz ne '' ) {
	    # Hmm. \bRe:\b seems not to match on "Re: "...
	    if ( $subj =~ /\b$re_buzz[\s\b]/i ) {
		notice ("Passed: Buzz word in subject");
		return 0;
	    }
	}
    }

    my $min = 5;
    my $max = 0;
    my $tag;

    # Get a list of all possible senders and verify them.
    my @addr = ();
    foreach $tag ( "From", "Resent-From", "Sender", "Reply-To" ) {
	next unless $head->count($tag);
	foreach ( $head->get($tag) ) {
	    my @o = Mail::Address->parse($_);
	    foreach ( @o ) {
		my $h = $_->address();
		next unless defined $h;
		if ( $tag eq "Sender" and $i_sender ne '' and 
		     lc($h) eq lc($i_sender) ) {
		    notice ("$tag: $h [Ignored]");
		    next;
		}
		push (@addr, $h);
		my $ret = vrfy_addr ($h);
		print STDERR ("Verify: $h -> $ret\n") if $debug;
		$min = $ret if $min > $ret;
		$max = $ret if $max < $ret;
		notice ("$tag: $h [".$reason[$ret]."]");
	    }
	}
    }

    if ( $max == 4 ) {
        # We do not even send a reply... it will probably bounce...
	notice ("Denied: Denied sender.");
	return 4;
    }

    if ( $min <= 1 ) {
	notice ("Passed: Known sender and no denied ones.");
	return 0;
    }

    if ( scalar (@addr) == 0 ) {
	notice ("Rejected: cannot determine sender address");
	return 3 unless $check ne '';
    }
    else {
	notice ("Rejected: suspicious or unknown sender");
    }

    # Remail it.
    my $key = sprintf ("%.8X", time - $offset);

    if ( $test ) {
#	open (MAILER, "|/bin/cat");
#	open (MAILER, ">> /home/felix/Mail/despam.messages");
	open (MAILER, "|tee -a /home/felix/Mail/despam.message | /usr/lib/sendmail -t");
    }
    else {
	open (MAILER, "|/usr/lib/sendmail -t");
    }
    print MAILER ($check ne '' ? "To: $check\nX-WouldSend-To: " : "To: ",
		  join(", ", @addr), "\n",
		  $from ne '' ? "From: $from\n" : '',
		  "Subject: REJECTED: ", $head->get("Subject"),
		  "X-Despammed-By: $despam_tag [$my_name $my_version]\n",
		  "\n");
    print MAILER <<EOD;
Madam, Sir,

Due to the dramatically increasing number of unsolicited and spam
mails I am getting in my mailbox, I am using an automatic mechanism
to filter out such mail.

Your message has been rejected by my despamming mail filter
program. Either because: 
    - you are a known spammer... (in which case you will probably not 
				  even read this) 
    - you are unknown to me... (most likely)

If you want your message to reach me, and thus make you known to my
despamming system, you have to add the word Z${key}V to the 
subject line and re-send it within two weeks from now. Alternatively,
you can add a mail header "X-Despam-Key: Z${key}V". However, by
resending the message this way you formally declare that your message
is not an unsolicited advertisement of some sort. Violating this will
make you eligible to legal actions.

My despamming system keeps track of the people which e-mail I have 
approved. Therefore, if you are not a spammer, you should only get this
message once.

If your message was not sent to me directly (but through a mail list
of some sort), I apologize for the inconvenience of bouncing it to
you. I invite you to resent it to me as described above and point this
problem to me.

        Johan Vromans <jvromans\@squirrel.nl>

EOD

    print MAILER ("-------",
		  " start of forwarded message (RFC 934 encapsulation) ",
		  "-------\n");
    print MAILER ($orig);

    while ( <> ) {
	s/^--/- --/;
	print MAILER ($_);
    }
    print MAILER ("-------",
		  " end ",
		  "-------\n");
    close (MAILER);

    return 3;
}

sub parse_data {
    my ($force) = @_;


    unless ( $force || !$cache ) {
	return if check_cache ("data", $datafile);
    }

    local (*BL);

    print STDERR ("=> opening $datafile\n") if $debug;
    open (BL, $datafile) || die ("$datafile: $!\n");

    $re_buzz = '(';
    $re_allow = '(';
    $re_deny = '(';
    $re_suspicious = '(';

    my $ref = undef;

    while ( <BL> ) {

	# Empty lines and lines starting with # are ignored.
	next if /^\s*#/;
	next unless /\S/;
	chomp;

	if    ( /^allow(\s|$)/ ) { $ref = \$re_allow; $_ = " " . $'; }
	elsif ( /^deny(\s|$)/ )  { $ref = \$re_deny;  $_ = " " . $'; }
	elsif ( /^suspicious(\s|$)/ )  { $ref = \$re_suspicious;  $_ = " " . $'; }
	elsif ( /^buzz(\s|$)/ )  { $ref = \$re_buzz;  $_ = " " . $'; }

	s/^\s+//;
	s/\s+$//;
	next unless /\S/;

	die ("Error in $datafile:\n".
	     "  missing \"allow\", \"deny\", \"suspicious\" or \"buzz\"\n")
		unless defined $$ref;

	# Lines starting with / denote a Perl regular expression pattern.
	if ( /^\// ) {
	    $$ref .= $' . "|";
	}

	# Other lines denote shell-type patterns.
	# Special case: xxx@@host.dom will match xxx@[anyhost.]host.dom .
	else {
	    while ( /(\*|\@\@)/ ) {
		$$ref .= quotemeta(lc($`));
		$$ref .= ($+ eq "*") ? ".*" : "\@(.+\\.)?";
		$_ = $';
	    }
	    $$ref .= quotemeta(lc($_)) . "|";
	}
    }

    close (BL);

    foreach $ref ( \$re_buzz, \$re_deny, \$re_suspicious, \$re_allow ) {
	if ( length ($$ref) > 1 ) {
	    substr($$ref, -1, 1) = ')';
	}
	else {
	    $$ref = "\n";	# should never match
	}
    }
    if ( $cache ) {
	write_cache ("data", $datafile);
    }
}

sub check_cache {
    my ($tag, $file) = @_;
    my $cache = $file . '.cache';
    my @a;
    my $cachetime;

    # Check for a cache file.
    if ( defined (@a = stat ($cache)) and  defined ($cachetime = $a[9])) {

	# Is it up to date?
	if ( defined (@a = stat ($file)) and $a[9] < $cachetime ) {

	    # Read its contents into the regexp.
	    print STDERR ("Using $tag cache $cache\n") if $verbose;

	    require $cache;
	    die ("Error reading $tag cache $cache\n")
		unless length ($re_deny) > 0 
		    and length ($re_suspicious) > 0 
		    and length ($re_allow) > 0 
		    and length ($re_buzz) > 0;
		if ( $debug ) {
		    print STDERR ("re_buzz  = '$re_buzz';\n");
		    print STDERR ("re_suspicious  = '$re_suspicious';\n");
		    print STDERR ("re_deny  = '$re_deny';\n");
		    print STDERR ("re_allow = '$re_allow';\n");
	    }

	    # And return it.
	    return 1;
	}
    }

    # No cache.
    return 0;
}

sub write_cache {
    my ($tag, $file) = @_;
    my $cache = $file . '.cache';
    local (*CF);

    # Create and write cache file.
    if ( open (CF, '>'.$cache) ) {
	print CF ("\$re_buzz = '$re_buzz';\n");
	print CF ("\$re_suspicious = '$re_suspicious';\n");
	print CF ("\$re_deny = '$re_deny';\n");
	print CF ("\$re_allow = '$re_allow';\n");

	close (CF);
	print STDERR ("Updated $tag cache $cache\n") if $verbose;
    }
    else {
	warn ("Error writing $tag cache $cache: $!\n");
    }
}

sub notice {
    my ($msg) = @_;
    print STDERR ("DESPAM: ", $msg, "\n");
}

sub options {
    my $help = 0;		# handled locally
    my $ident = 0;		# handled locally
    my $man = 0;		# handled locally

    # Process options.
    if ( @ARGV > 0 and $ARGV[0] =~ /^[-+]/ ) {
	GetOptions ('ident' => \$ident,
		    'addresses=s' => \$addresses,
		    'data=s' => \$datafile,
		    'ignore_sender=s' => \$i_sender,
		    'domain=s' => \$domain,
		    'tag=s' => \$despam_tag,
		    'from=s' => \$from,
		    'key=i' => \$offset,
		    'check=s' => \$check,
		    'verify' => \$verify,
		    'verbose' => \$verbose,
		    'trace' => \$trace,
		    'help' => \$help,
		    'man' => \$man,
		    'test' => \$test,
		    'debug' => \$debug);
    }
    if ( $ident or $help or $man ) {
	print STDERR ("This is $my_package [$my_name $my_version]\n");
    }
    if ( $help ) {
	# # Load Pod::Usage only if needed.
	# if ( do "Pod/Usage.pm" ) {
	#     import Pod::Usage;
	#     pod2usage (1);
	# }
	# else {
	    usage ();
	# }
    }
    elsif ( $man ) {
	# Load Pod::Usage only if needed.
	require "Pod/Usage.pm";
	import Pod::Usage;
	pod2usage (VERBOSE => 2);
    }
}

sub usage {
    print STDERR <<EndOfUsage;
Usage: $0 -verify [options] addresses...
Usage: $0 [options] < file
    -addresses XXX	file with list of email adresses (e.g. from BBDB)
    -data XXX		file with buzzwords and allowed/denied addresses
    -ignore_sender XXX	ignore this sender in Sender: headers
    -domain XXX		allow more from this domain
    -tag XXX		despamming tag 
    -from XXX		use this a From: address
    -key NNN		starting key for calculations
    -[no]cache		[do not] cache the data file
    -check XXX		remail to this address instead of to the senders
    -help		this message
    -man		prints the manual
    -ident		show identification
    -verbose		verbose information
EndOfUsage
    exit 1;
}
!NO!SUBS!
close OUT or die "Can't close $file: $!";
chmod 0755, $file or die "Can't reset permissions for $file: $!\n";
exec("$Config{'eunicefix'} $file") if $Config{'eunicefix'} ne ':';
